from django.http import JsonResponse
from django.shortcuts import render
from django.urls import reverse


def get_swagger_json(request, router):
    return JsonResponse(router.api)


def render_docs(request, router, json_url_name):
    return render(request, 'lepo_doc/swagger-ui.html', {
        'json_url': request.build_absolute_uri(reverse(json_url_name)),
    })



from collections import OrderedDict

from django.utils.functional import cached_property

from lepo.utils import maybe_resolve


class Operation:
    def __init__(self, api, path, method, data):
        """
        :type api: lepo.router.Router
        :type path: lepo.path.Path
        :type method: str
        :type data: dict
        """
        self.api = api
        self.path = path
        self.method = method
        self.data = data

    @property
    def id(self):
        return self.data['operationId']

    @cached_property
    def parameters(self):
        """
        Combined path-level and operation-level parameters.

        Any $refs are resolved here.

        Note that this implementation differs from the spec in that we only use
        the _name_ of a parameter to consider its uniqueness, not the name and location.

        This is because we end up passing parameters to the handler by name anyway,
        so any duplicate names, even if they had different locations, would be horribly mangled.

        :rtype: list[dict]
        """

        parameters = OrderedDict()
        for source in (
            self.path.mapping.get('parameters', ()),
            self.data.get('parameters', {}),
        ):
            source = maybe_resolve(source, self.api.resolve_reference)
            for parameter in source:
                parameter = maybe_resolve(parameter, self.api.resolve_reference)
                parameters[parameter['name']] = parameter

        return list(parameters.values())

    def _get_overridable(self, key, default=None):
        # TODO: This probes a little too deeply into the specifics of these objects, I think...
        for obj in (
            self.data,
            self.path.mapping,
            self.api.api,
        ):
            if key in obj:
                return obj[key]
        return default

    @cached_property
    def consumes(self):
        value = self._get_overridable('consumes', [])
        if not isinstance(value, (list, tuple)):
            raise TypeError('`consumes` must be a list, got %r' % value)  # pragma: no cover
        return value

    @cached_property
    def produces(self):
        value = self._get_overridable('produces', [])
        if not isinstance(value, (list, tuple)):
            raise TypeError('`produces` must be a list, got %r' % value)  # pragma: no cover
        return value



import argparse
import sys

import yaml
from six import StringIO
from django.utils.text import camel_case_to_spaces

from lepo.router import Router

HANDLER_TEMPLATE = '''
def {func_name}(request, {parameters}):
    raise NotImplementedError('Handler {operation_id} not implemented')
'''.strip()


def generate_handler_stub(router, handler_template=HANDLER_TEMPLATE):
    output = StringIO()
    func_name_to_operation = {}
    for path in router.get_paths():
        for operation in path.get_operations():
            snake_operation_id = camel_case_to_spaces(operation.id).replace(' ', '_')
            func_name_to_operation[snake_operation_id] = operation
    for func_name, operation in sorted(func_name_to_operation.items()):
        parameter_names = [p['name'] for p in operation.parameters]
        handler = handler_template.format(
            func_name=func_name,
            operation_id=operation.id,
            parameters=', '.join(parameter_names),
        )
        output.write(handler)
        output.write('\n\n\n')
    return output.getvalue()


def cmdline():
    ap = argparse.ArgumentParser()
    ap.add_argument('input', default=None, nargs='?')
    args = ap.parse_args()
    input = (open(args.input) if args.input else sys.stdin)
    api = yaml.safe_load(input)
    print(generate_handler_stub(Router(api)))


if __name__ == '__main__':
    cmdline()



import re

from lepo.excs import InvalidOperation
from lepo.operation import Operation
from lepo.path_view import PathView

# As defined in the documentation for Path Items:
METHODS = {'get', 'put', 'post', 'delete', 'options', 'head', 'patch'}


class Path:
    def __init__(self, api, path, mapping):
        self.api = api
        self.path = path
        self.mapping = mapping
        self.regex = re.sub(
            r'\{(.+?)\}',
            lambda m: '(?P<%s>.+?)' % m.group(1),
            self.path,
        ).lstrip('/') + '$'
        self.name = re.sub(  # todo: embetter
            r'[^a-z0-9]+',
            '-',
            self.path,
            re.I,
        ).strip('-')
        self.view_class = type('%sView' % self.name.title(), (PathView,), {
            'path': self,
            'api': self.api,
        })

    def get_operation(self, method):
        operation_data = self.mapping.get(method.lower())
        if not operation_data:
            raise InvalidOperation('Path %s does not support method %s' % (self.path, method.upper()))
        return Operation(api=self.api, path=self, data=operation_data, method=method)

    def get_operations(self):
        for method in METHODS:
            if method in self.mapping:
                yield self.get_operation(method)



from copy import deepcopy
from functools import wraps

from django.utils.module_loading import import_string


class BaseHandler:
    def __init__(self, request, args):
        self.request = request
        self.args = args

    @classmethod
    def get_view(cls, method_name):
        """
        Get a Django function view calling the given method (and pre/post-processors)

        :param method_name: The method on the class
        :return: View function
        """
        method = getattr(cls, method_name)

        @wraps(method)
        def view(request, **kwargs):
            handler = cls(request, kwargs)
            handler.call_processors('view')
            response = None
            try:
                response = method(handler)
                return response
            finally:
                handler.call_processors('post_view', response=response)

        return view

    def get_processors(self, purpose):
        return getattr(self, '%s_processors' % purpose, ())

    def call_processors(self, purpose, **kwargs):
        for proc in self.get_processors(purpose):
            if isinstance(proc, str):
                proc = getattr(self, proc, None) or import_string(proc)
            kwargs['purpose'] = purpose
            proc(**kwargs)


class BaseModelHandler(BaseHandler):
    model = None
    queryset = None
    schema_class = None

    id_data_name = 'id'
    id_field_name = 'pk'

    def get_schema(self, purpose, object=None):
        schema_class = getattr(self, '%s_schema_class' % purpose, None)
        if schema_class is None:
            schema_class = self.schema_class
        kwargs = {}
        if purpose == 'update':
            kwargs['partial'] = True
        return schema_class(**kwargs)

    def get_queryset(self, purpose):
        queryset = getattr(self, '%s_queryset' % purpose, None)
        if queryset is None:
            queryset = self.queryset
        return deepcopy(queryset)

    def process_object_list(self, purpose, object_list):
        return object_list

    def retrieve_object(self):
        queryset = self.get_queryset('retrieve')
        object = queryset.get(**{self.id_field_name: self.args[self.id_data_name]})
        self.call_processors('retrieve_object', object=object)
        return object


class ModelHandlerReadMixin(BaseModelHandler):
    def handle_list(self):
        self.call_processors('list')
        queryset = self.get_queryset('list')
        object_list = self.process_object_list('list', queryset)
        schema = self.get_schema('list')
        return schema.dump(object_list, many=True).data

    def handle_retrieve(self):
        self.call_processors('retrieve')
        object = self.retrieve_object()
        schema = self.get_schema('retrieve')
        return schema.dump(object).data


class ModelHandlerCreateMixin(BaseModelHandler):
    create_data_name = 'data'

    def handle_create(self):
        self.call_processors('create')
        schema = self.get_schema('create')
        result = schema.load(self.args[self.create_data_name])
        assert not result.errors
        data = result.data
        if not isinstance(data, self.model):
            data = self.model(**data)

        data.full_clean()
        data.save()

        schema = self.get_schema('post_create', object=data)
        return schema.dump(data).data


class ModelHandlerUpdateMixin(BaseModelHandler):
    update_data_name = 'data'

    def handle_update(self):
        self.call_processors('update')
        object = self.retrieve_object()
        schema = self.get_schema('update', object=object)
        result = schema.load(self.args[self.update_data_name])
        assert not result.errors
        for key, value in result.data.items():
            setattr(object, key, value)
        object.full_clean()
        object.save()
        self.call_processors('post_update', object=object)
        schema = self.get_schema('post_update', object=object)
        return schema.dump(object).data


class ModelHandlerDeleteMixin(BaseModelHandler):
    def handle_delete(self):
        self.call_processors('delete')
        object = self.retrieve_object()
        object.delete()
        self.call_processors('post_delete', object=object)


class CRUDModelHandler(
    ModelHandlerCreateMixin,
    ModelHandlerReadMixin,
    ModelHandlerUpdateMixin,
    ModelHandlerDeleteMixin,
):
    pass



class APIInfo:
    def __init__(self, operation):
        """
        :type operation: lepo.operation.Operation
        """
        self.api = operation.api
        self.path = operation.path
        self.operation = operation



def maybe_resolve(object, resolve):
    """
    Call `resolve` on the `object`'s `$ref` value if it has one.

    :param object: An object.
    :param resolve: A resolving function.
    :return: An object, or some other object! :sparkles:
    """
    if isinstance(object, dict) and object.get('$ref'):
        return resolve(object['$ref'])
    return object



import base64
import json

import iso8601
import jsonschema
from django.utils.encoding import force_bytes, force_text

from lepo.excs import ErroneousParameters, InvalidBodyContent, InvalidBodyFormat, MissingParameter
from lepo.utils import maybe_resolve

COLLECTION_FORMAT_SPLITTERS = {
    'csv': lambda value: force_text(value).split(','),
    'ssv': lambda value: force_text(value).split(' '),
    'tsv': lambda value: force_text(value).split('\t'),
    'pipes': lambda value: force_text(value).split('|'),
}

OPENAPI_JSONSCHEMA_VALIDATION_KEYS = (
    'maximum', 'exclusiveMaximum',
    'minimum', 'exclusiveMinimum',
    'maxLength', 'minLength',
    'pattern',
    'maxItems', 'minItems',
    'uniqueItems',
    'enum', 'multipleOf',
)


def cast_parameter_value(api_info, parameter, value):
    if parameter.get('type') == 'array':
        if not isinstance(value, list):  # could be a list already if collection format was multi
            splitter = COLLECTION_FORMAT_SPLITTERS.get(parameter.get('collectionFormat', 'csv'))
            if not splitter:
                raise NotImplementedError('unsupported collection format in %r' % parameter)
            value = splitter(value)
        items = parameter['items']
        value = [cast_parameter_value(api_info, items, item) for item in value]
    if 'schema' in parameter:
        schema = maybe_resolve(parameter['schema'], api_info.api.resolve_reference)
        jsonschema.validate(value, schema, resolver=api_info.api.resolver)
        if 'discriminator' in schema:  # Swagger Polymorphism support
            type = value[schema['discriminator']]
            actual_type = '#/definitions/%s' % type
            schema = api_info.api.resolve_reference(actual_type)
            jsonschema.validate(value, schema, resolver=api_info.api.resolver)
        return value
    value = cast_primitive_value(parameter, value)
    jsonschema_validation_object = {
        key: parameter[key]
        for key in parameter
        if key in OPENAPI_JSONSCHEMA_VALIDATION_KEYS
    }
    if jsonschema_validation_object:
        jsonschema.validate(value, jsonschema_validation_object)
    return value


def cast_primitive_value(spec, value):
    format = spec.get('format')
    type = spec.get('type')
    if type == 'boolean':
        return (force_text(value).lower() in ('1', 'yes', 'true'))
    if type == 'integer' or format in ('integer', 'long'):
        return int(value)
    if type == 'number' or format in ('float', 'double'):
        return float(value)
    if format == 'byte':  # base64 encoded characters
        return base64.b64decode(value)
    if format == 'binary':  # any sequence of octets
        return force_bytes(value)
    if format == 'date':  # ISO8601 date
        return iso8601.parse_date(value).date()
    if format == 'dateTime':  # ISO8601 datetime
        return iso8601.parse_date(value)
    if type == 'string':
        return force_text(value)
    return value


def read_body(request):
    consumes = request.api_info.operation.consumes
    if request.content_type not in consumes:
        raise InvalidBodyFormat('Content-type %s is not supported (%r are)' % (
            request.content_type,
            consumes,
        ))
    try:
        if request.content_type == 'application/json':
            return json.loads(request.body.decode(request.content_params.get('charset', 'UTF-8')))
        elif request.content_type == 'text/plain':
            return request.body.decode(request.content_params.get('charset', 'UTF-8'))
    except Exception as exc:
        raise InvalidBodyContent('Unable to parse this body as %s' % request.content_type) from exc
    raise NotImplementedError('No idea how to parse content-type %s' % request.content_type)  # pragma: no cover


def get_parameter_value(request, view_kwargs, param):
    if param['in'] == 'formData' and param.get('type') == 'file':
        return request.FILES[param['name']]

    if param['in'] in ('query', 'formData'):
        source = (request.POST if param['in'] == 'formData' else request.GET)
        if param.get('type') == 'array' and param.get('collectionFormat') == 'multi':
            return source.getlist(param['name'])
        else:
            return source[param['name']]

    if param['in'] == 'path':
        return view_kwargs[param['name']]

    if param['in'] == 'body':
        return read_body(request)

    if param['in'] == 'header':
        return request.META['HTTP_%s' % param['name'].upper().replace('-', '_')]

    raise NotImplementedError('unsupported `in` value in %r' % param)  # pragma: no cover


def read_parameters(request, view_kwargs):
    """
    :param request: HttpRequest with attached api_info
    :type request: HttpRequest
    :type view_kwargs: dict[str, object]
    :rtype: dict[str, object]
    """
    params = {}
    errors = {}
    for param in request.api_info.operation.parameters:
        try:
            value = get_parameter_value(request, view_kwargs, param)
        except KeyError:
            if 'default' in param:
                params[param['name']] = param['default']
                continue
            if param.get('required'):  # Required but missing
                errors[param['name']] = MissingParameter('parameter %s is required but missing' % param['name'])
            continue
        try:
            params[param['name']] = cast_parameter_value(request.api_info, param, value)
        except NotImplementedError:
            raise
        except Exception as e:
            errors[param['name']] = e
    if errors:
        raise ErroneousParameters(errors, params)
    return params



class MissingParameter(ValueError):
    pass


class InvalidOperation(ValueError):
    pass


class MissingHandler(ValueError):
    pass


class ErroneousParameters(Exception):
    def __init__(self, error_map, parameters):
        self.errors = error_map
        self.parameters = parameters


class InvalidBodyFormat(ValueError):
    pass


class InvalidBodyContent(ValueError):
    pass


class RouterValidationError(Exception):
    def __init__(self, error_map):
        self.errors = error_map
        self.description = '\n'.join('%s: %s' % (key, value) for (key, value) in sorted(self.errors.items()))
        super(RouterValidationError, self).__init__('Router validation failed:\n%s' % self.description)



from copy import deepcopy
from importlib import import_module

from django.conf.urls import url
from django.utils.text import camel_case_to_spaces
from jsonschema import RefResolver

from lepo.excs import MissingHandler
from lepo.path import Path
from lepo.utils import maybe_resolve


class Router:
    path_class = Path

    def __init__(self, api):
        self.api = deepcopy(api)
        self.api.pop('host', None)
        self.handlers = {}
        self.resolver = RefResolver('', self.api)

    @classmethod
    def from_file(cls, filename):
        with open(filename) as infp:
            if filename.endswith('.yaml') or filename.endswith('.yml'):
                import yaml
                data = yaml.safe_load(infp)
            else:
                import json
                data = json.load(infp)
        return cls(data)

    def get_path(self, path):
        """
        Construct a Path object from a path string.

        The Path string must be declared in the API.

        :type path: str
        :rtype: lepo.path.Path
        """
        mapping = maybe_resolve(self.api['paths'][path], self.resolve_reference)
        return self.path_class(api=self, path=path, mapping=mapping)

    def get_paths(self):
        for path in self.api['paths']:
            yield self.get_path(path)

    def get_urls(self):
        urls = []
        for path in self.get_paths():
            urls.append(url(path.regex, path.view_class.as_view(), name=path.name))
        return urls

    def get_handler(self, operation_id):
        if operation_id in self.handlers:
            return self.handlers[operation_id]
        snake_operation_id = camel_case_to_spaces(operation_id).replace(' ', '_')
        if snake_operation_id in self.handlers:
            return self.handlers[snake_operation_id]
        raise MissingHandler(
            'Missing handler for operation %s (tried %s too)' % (operation_id, snake_operation_id)
        )

    def add_handlers(self, namespace):
        if isinstance(namespace, str):
            namespace = import_module(namespace)
        for name, value in vars(namespace).items():
            if name.startswith('_'):
                continue
            try:
                if callable(value):
                    self.handlers[name] = value
            except:
                pass

    def resolve_reference(self, ref):
        url, resolved = self.resolver.resolve(ref)
        return resolved



from django.http import HttpResponse, JsonResponse
from django.views import View

from lepo.api_info import APIInfo
from lepo.excs import InvalidOperation
from lepo.parameters import read_parameters


class PathView(View):
    api = None  # Filled in by subclasses
    path = None  # Filled in by subclasses

    def dispatch(self, request, **kwargs):
        try:
            operation = self.path.get_operation(request.method)
        except InvalidOperation:
            return self.http_method_not_allowed(request, **kwargs)
        request.api_info = APIInfo(operation=operation)
        params = read_parameters(request, kwargs)
        handler = request.api_info.api.get_handler(operation.id)
        response = handler(request, **params)
        if isinstance(response, HttpResponse):
            # TODO: validate against responses
            return response
        return JsonResponse(response, safe=False)  # TODO: maybe less TIMTOWDI here?



from django.conf.urls import url


def get_docs_urls(router, namespace, docs_url='docs/?'):
    from . import views
    json_url_name = 'lepo_doc_%s' % id(router)
    return [
        url('swagger\.json$', views.get_swagger_json, kwargs={'router': router}, name=json_url_name),
        url('%s$' % docs_url, views.render_docs, kwargs={
            'router': router,
            'json_url_name': '%s:%s' % (namespace, json_url_name),
        }),
    ]



from lepo.excs import RouterValidationError


def validate_router(router):
    errors = {}
    operations = set()
    for path in router.get_paths():
        for operation in path.get_operations():
            operations.add(operation.id)
    for operation in operations:
        try:
            router.get_handler(operation)
        except Exception as e:
            errors[operation] = e
    if errors:
        raise RouterValidationError(errors)



import setuptools

if __name__ == '__main__':
    setuptools.setup(
        name='lepo',
        version='0.0.0',
        url='https://github.com/akx/lepo',
        author='Aarni Koskela',
        author_email='akx@iki.fi',
        maintainer='Aarni Koskela',
        maintainer_email='akx@iki.fi',
        license='MIT',
        install_requires=['Django', 'iso8601', 'jsonschema', 'marshmallow'],
        packages=setuptools.find_packages('.', exclude=(
            'lepo_tests',
            'lepo_tests.*',
        )),
        include_package_data=True,
    )
